import datetime as dt
import sys
import time
import typing as t

import dateparser


class Zapper:
    """
    Wait for a) pressing enter, or b) a few seconds.
    """

    def __init__(self, when: t.Optional[str] = None, action: t.Optional[t.Callable] = None) -> None:
        self.when = when or ""
        self.action: t.Optional[t.Callable] = action
        self.validate()
        self.delay: float = self._compute_delay()

    @property
    def is_stopclock(self):
        return self.when.endswith("s")

    @property
    def is_keypress(self):
        return self.when.startswith("key")

    def _compute_delay(self) -> float:
        if self.is_stopclock:
            duration = dateparser.parse(self.when)
            if duration is None:
                raise ValueError(f"Unable to parse duration: {duration}")
            return (dt.datetime.now() - duration).total_seconds()
        return 0

    def validate(self) -> bool:
        if self.when == "" or self.is_stopclock or self.is_keypress:
            return True
        raise ValueError(
            f"Invalid value for `when` argument: {self.when}. "
            f"Either provide a duration (e.g. 1.42s), or `keypress`."
        )

    def process(self) -> bool:
        if self.when:
            self.wait()
            if self.action is not None:
                self.action()
            return True
        return False

    def wait(self) -> None:
        if self.is_stopclock:
            time.sleep(self.delay)
        elif self.is_keypress:
            print("Press enter to zap and continue the program flow.", file=sys.stderr)
            tty_input()


def tty_input(prompt: t.Union[str, None] = None):
    """
    Read string from TTY.
    """
    with open("/dev/tty") as terminal:
        sys.stdin = terminal
        if prompt is None:
            user_input = input()
        else:
            user_input = input(prompt)
    sys.stdin = sys.__stdin__
    return user_input



import click

github_organization_option = click.option(
    "--github-organization", "--gh-org", type=str, required=True, help="GitHub organization name"
)
github_repository_option = click.option(
    "--github-repository",
    "--gh-repo",
    type=str,
    required=False,
    help="GitHub repository, single or path to file",
)
format_option = click.option(
    "--format",
    "format_",
    type=str,
    required=False,
    default="markdown",
    help="Output format. Default: markdown",
)
when_option = click.option("--when", type=str, required=False, help="Point in time")



"""
A small wrapper around `github-backup`.
https://pypi.org/project/github-backup/
"""

import subprocess
from shutil import which


class GitHubBackup:
    def run(self, args):
        cmd = [which("github-backup"), *args]
        subprocess.check_call(cmd)  # noqa: S603



import datetime
import os
from io import StringIO
from pathlib import Path

project_blacklist = ["gitstats"]


class Project:
    def __init__(self, name="", path="", vcs=""):
        self.name = name
        self.path = path
        self.vcs = vcs
        self.changes_file = None

    def __str__(self):
        return "Project={0}".format(self.name)

    def __repr__(self):
        return str(self)


def mkproject(path: str):
    if os.path.isdir(path):
        subdirs = os.listdir(path)
        if ".svn" in subdirs:
            vcs = "svn"
        elif ".git" in subdirs:
            vcs = "git"
        else:
            raise FileNotFoundError(f"Project is not a VCS: {path}")
        p = Path(path).absolute()
        project = Project(p.name, path, vcs)
        return project
    else:
        raise FileNotFoundError(f"Project is not a path: {path}")


def walk_projects(source_directory):
    """Scan all main level directories in the projects directory"""
    for project_name in sorted(os.listdir(source_directory)):
        if project_name in project_blacklist:
            continue
        project_path = os.path.join(source_directory, project_name)
        yield mkproject(project_path)


def now():
    return datetime.datetime.now().strftime("%Y-%m-%d %H:%M")


def rest_header(title="", generator=""):
    buffer = StringIO()
    buffer.write(".. note:: Generated by ``{0}`` on {1}".format(generator, now()))
    buffer.write("\n\n")
    buffer.write("=" * len(title) + "\n")
    buffer.write(title + "\n")
    buffer.write("=" * len(title) + "\n")
    buffer.write("\n")
    buffer.seek(0)
    return buffer.read()



import typing as t

import click

from rapporto.source.changes.core import aggregate


@click.command(context_settings=dict(ignore_unknown_options=True))
@click.argument("input", type=click.UNPROCESSED, required=False, nargs=-1)
@click.option("--output", type=str, required=False, help="Output path")
@click.option(
    "--format",
    "format_",
    type=click.Choice(["md", "rst", "text"]),
    required=False,
    default="rst",
    help="Output format: 'md' for markdown, 'rst' for reStructuredText, 'text' for terminal output",
)
@click.pass_context
def cli(ctx: click.Context, input: t.List[str], output: str, format_: str) -> None:  # noqa: A002
    """
    Aggregate change log files.
    """
    aggregate(input[0], output)



import datetime as dt
import logging
import sys
import typing as t

from aika import TimeIntervalParser
from markdown_to_mrkdwn import SlackMarkdownConverter

logger = logging.getLogger(__name__)


def sanitize_title(title: str) -> str:
    """
    Strip characters that are unfortunate in Markdown link titles.
    """
    return title.replace("[", "⎡").replace("]", "⎦")


def goosefeet(text: str) -> str:
    """
    Parameters including spaces need to be treated specially for the GitHub API.
    """
    if " " in text:
        return f'"{text}"'
    return text


def to_mrkdwn(markdown: str, unordered_list: bool = True) -> str:
    """
    Convert standard Markdown to Slack `mrkdwn` format.
    """
    mrkdwn_converter = SlackMarkdownConverter()
    try:
        # Unordered list
        if not unordered_list:
            mrkdwn_converter.patterns.remove((r"^(\s*)- (.+)", r"\1• \2"))
    except ValueError:
        pass
    return mrkdwn_converter.convert(markdown)


def setup_logging(level=logging.INFO, verbose: bool = False):
    """
    Configure Python's logging module.
    """
    log_format = "%(asctime)-15s [%(name)-36s] %(levelname)-8s: %(message)s"
    logging.basicConfig(format=log_format, stream=sys.stderr, level=level)


def week_to_day_range(when: str, skip_the_future: bool = True) -> t.List[str]:
    """
    From a current point in time, derive the calendar week in ISO8601 format.

    TODO: Refactor to Aika.
    TODO: Make it configurable whether to return 7 or 8 days.
    """
    week = []
    today = dt.date.today()
    tip = TimeIntervalParser()
    interval = tip.parse(when)
    cursor = interval.start
    while cursor < interval.end:
        week.append(cursor.strftime("%Y-%m-%d"))
        cursor += dt.timedelta(days=1)
        if skip_the_future and cursor.date() > today:
            logger.info(f"Skipping day in the future: {cursor.date().isoformat()}.")
            break
    return week



import typing as t

import attr
import click
from attrs import define
from furl import furl

slack_api_token_option = click.option(
    "--slack-token", type=str, envvar="SLACK_TOKEN", required=False, help="Slack API token"
)
slack_channel_option = click.option(
    "--slack-channel", "-c", type=str, envvar="SLACK_CHANNEL", required=False, help="Slack channel"
)


@define
class SlackOptions:
    """
    Composite Slack options.
    """

    token: str
    channel: str


@define
class SlackUrl:
    """
    Manage and decode a full HTTP Slack URL.

    https://acme.slack.com/archives/C08EF2NGZGB
    https://acme.slack.com/archives/C08EF2NGZGB/p1740421750904349
    https://acme.slack.com/archives/C08EF2NGZGB/p1740478361323219?thread_ts=1740421750.904349&cid=C08EF2NGZGB
    """

    url: str
    _url: furl = attr.field(init=False, default=None)

    def __attrs_post_init__(self):
        self._url = furl(self.url)

    @classmethod
    def from_url(cls, url: str) -> "SlackUrl":
        return cls(url=url)

    @property
    def channel_id(self) -> str:
        return self._url.path.segments[1]

    @property
    def ts(self) -> t.Union[str, None]:
        try:
            message_id = self._url.path.segments[2].lstrip("p")
            message_id = message_id[:-6] + "." + message_id[-6:]
            return message_id
        except IndexError:
            return None

    @property
    def cid(self) -> str:
        return self._url.query.params.get("cid")

    @property
    def thread_ts(self) -> str:
        return self._url.query.params.get("thread_ts")


class SlackChannel:
    @classmethod
    def from_any(cls, value: t.Union[str, None]) -> t.Union[str, None]:
        if value is None:
            return None
        if value.startswith("http://") or value.startswith("https://"):
            return SlackUrl.from_url(value).channel_id
        else:
            return value.lstrip("#")


class SlackMessage:
    @classmethod
    def from_any(cls, value: t.Union[str, None]) -> t.Union[str, None]:
        if value is None:
            return None
        if value.startswith("http://") or value.startswith("https://"):
            return SlackUrl.from_url(value).ts
        else:
            return value



"""
https://api.slack.com/methods/chat.postMessage
https://api.slack.com/surfaces/messages#threading
https://api.slack.com/messaging/retrieving#finding_conversation
https://api.slack.com/methods/conversations.history
https://api.slack.com/methods/conversations.replies
"""

import logging
import typing as t
import warnings
from contextlib import contextmanager

from munch import Munch, munchify
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError

from pueblo_goof.slack.model import SlackChannel, SlackMessage, SlackOptions

logger = logging.getLogger(__name__)


class SlackConversation:
    """
    Wrap a Slack conversation.
    """

    def __init__(
        self,
        options: t.Optional[SlackOptions] = None,
        api_token: t.Optional[str] = None,
        channel: t.Optional[str] = None,
    ):
        self.options = options or SlackOptions(token=api_token, channel=channel)
        self.api_token = self.options.token
        self.message_ids: t.List[str] = []
        self.webclient = WebClient(token=self.api_token)
        self.channel_id = self.decode_channel(self.options.channel)

    def decode_channel(self, channel: str) -> str:
        """
        Decode channel id from channel id, name, or URL.
        """
        channel_search = SlackChannel.from_any(channel)
        try:
            return self.find_channel(what=channel_search).id
        except (AttributeError, KeyError, TypeError) as e:
            raise KeyError(f"Resolving channel failed: {channel_search}") from e

    def channels(self, limit: int = 999) -> t.List[Munch]:
        """
        Enumerate all channels.
        """
        return munchify(self.webclient.conversations_list(limit=limit)["channels"])

    def find_channel(self, what: t.Optional[str] = None):
        """
        Find channel by id or name.
        """
        for channel in self.channels():
            # print("channel:", channel)
            if channel.id == what or channel.name == what:
                return channel
        raise KeyError(f"Unable to find channel: {what}")

    def messages(self, limit: int = 20):
        """
        Enumerate all messages.
        """
        return munchify(
            self.webclient.conversations_history(
                channel=self.channel_id,
                limit=limit,
                inclusive=True,
                include_all_metadata=True,
            )["messages"]
        )

    def replies(self, ts: str, limit: int = 20):
        """
        Enumerate all replies.

        https://api.slack.com/methods/conversations.replies
        """
        return munchify(
            self.webclient.conversations_replies(
                channel=self.channel_id,
                ts=ts,
                limit=limit,
                inclusive=True,
                include_all_metadata=True,
            )["messages"]
        )

    def find_message_by_text(self, *labels):
        """
        Find message by labels in text body.
        """
        for message in self.messages():
            text = message["text"]
            if all(label in text for label in labels):
                return message
        return None

    @staticmethod
    def find_message_by_metadata(messages, **metadata):
        """
        Find message by metadata information.

        https://api.slack.com/reference/metadata
        """
        for message in messages:
            message_metadata = message.get("metadata", {}).get("event_payload", {})
            if not message_metadata:
                continue
            found = True
            for key, value in metadata.items():
                if message_metadata.get(key, None) != value:
                    found = False
                    break
            if found:
                return message
        return None

    def send(
        self,
        markdown: t.Union[str, None] = None,
        mrkdwn: t.Union[str, None] = None,
        reply_to: t.Union[str, None] = None,
        event: t.Union[str, None] = None,
        metadata: t.Union[t.Dict[str, str], None] = None,
    ):
        """
        Submit new Slack message to channel.
        """
        message_options = self._get_message_options(markdown=markdown, mrkdwn=mrkdwn)

        reply_to = SlackMessage.from_any(reply_to)

        try:
            # Ephemeral messages can be deleted by the addressed user,
            # but on the other hand, can't be deleted by the bot.
            """
            response = self._webclient.chat_postEphemeral(
                channel=self.channel,
                text=mrkdwn,
                type="mrkdwn",
                thread_ts=reply_to,
                user="U017YMHL5FY",
            )
            response["message_ts"]
            """
            metadata_effective = None
            if event and metadata:
                metadata_effective = {
                    "event_type": event,
                    "event_payload": metadata,
                }
            with self._suppress_top_level_text_warning():
                response = self.webclient.chat_postMessage(
                    channel=self.channel_id,
                    thread_ts=reply_to,
                    **message_options,
                    metadata=metadata_effective,
                    # icon_emoji=":chart_with_upwards_trend:",
                )
            logger.info(f"Sent message: {response['ts']}")
            self.message_ids.append(response["ts"])
            return response
        except SlackApiError as e:
            logger.error(f"{e.response['error']}. Status code: {e.response.status_code}")

    def update(
        self,
        ts: str,
        markdown: t.Union[str, None] = None,
        mrkdwn: t.Union[str, None] = None,
        event: t.Union[str, None] = None,
        metadata: t.Union[t.Dict[str, str], None] = None,
    ):
        """
        Update existing Slack message.
        """
        message_options = self._get_message_options(markdown=markdown, mrkdwn=mrkdwn)
        ts_effective = SlackMessage.from_any(ts)
        if ts_effective is None:
            raise ValueError(f"Unable to decode message id: {ts}")
        metadata_effective = None
        if event and metadata:
            metadata_effective = {
                "event_type": event,
                "event_payload": metadata,
            }

        with self._suppress_top_level_text_warning():
            response = self.webclient.chat_update(
                channel=self.channel_id,
                ts=ts_effective,
                **message_options,
                metadata=metadata_effective,
            )

        logger.info(f"Updated message: {response['ts']}")
        return response

    @contextmanager
    def _suppress_top_level_text_warning(self):
        """
        Suppress `UserWarning` emitted by Python Slack API.

        UserWarning: The top-level `text` argument is missing in the request payload for a
                     chat.update call - It's a best practice to always provide a `text` argument
                     when posting a message. The `text` argument is used in places where content
                     cannot be rendered such as: system push notifications, assistive technology
                     such as screen readers, etc.
          warnings.warn(ww, UserWarning)

        https://github.com/tech-writing/rapporto/issues/29
        """
        effective_warnings = []
        with warnings.catch_warnings(record=True, category=UserWarning) as catched_warnings:
            warnings.simplefilter("always")
            yield
            for raw_item in catched_warnings:
                if "top-level `text` argument is missing" not in str(raw_item.message):
                    effective_warnings.append(raw_item.message)

        for warning_item in effective_warnings:
            warnings.warn(warning_item, UserWarning, stacklevel=1)

    def _get_message_options(
        self, markdown: t.Union[str, None] = None, mrkdwn: t.Union[str, None] = None
    ):
        """
        Get message body options, either for `mrkdwn` or `markdown`.
        """
        # if markdown and not mrkdwn:
        #    mrkdwn = to_mrkdwn(markdown, unordered_list=False)
        message_options: t.Dict[str, t.Any] = {
            "unfurl_links": False,
            "unfurl_media": False,
        }

        # Vanilla Markdown
        if markdown:
            message_options["markdown_text"] = markdown
        # Slack mrkdwn
        elif mrkdwn:
            message_options["text"] = mrkdwn

        """
        message_options = {}
        message_options["blocks"] = [
            {
                "type": "section",
                "text": {"type": "mrkdwn", "text": mrkdwn},
                #"text": {"type": "markdown", "text": markdown},
                # "label": {"type": "plain_text", "text": "Feedback", "emoji": True},
            }
        ]
        """

        return message_options

    def delete(self):
        """
        Remove entire conversation, i.e. all replies.
        """
        message_ids = list(reversed(self.message_ids))
        logger.info(f"Deleting messages: {message_ids}")
        for message_id in message_ids:
            self.delete_message(message_id)

    def delete_message(self, identifier: str):
        """
        Delete specific message by identifier.
        """
        identifier_effective = SlackMessage.from_any(identifier)
        if identifier_effective is None:
            raise ValueError(f"Unable to decode message id: {identifier}")
        response = self.webclient.chat_delete(channel=self.channel_id, ts=identifier_effective)
        logger.info(f"Deleted message: {response['ts']}")
        return response

    def get_permalink(self, message_id: str) -> str:
        """
        Retrieve a permalink URL for a specific message.

        https://api.slack.com/methods/chat.getPermalink
        """
        response = self.webclient.chat_getPermalink(channel=self.channel_id, message_ts=message_id)
        return response.data["permalink"]



"""
OPSGENIE ALERTS REPORT GENERATOR

Generate report of Opsgenie alerts within a given time range.
Output formats can be Markdown or nicely formatted terminal output.

Usage Example:
    export OPSGENIE_API_KEY="your-api-key"
    rapporto opsgenie export-alerts --when="-7d"
    rapporto opsgenie export-alerts --start-time "12-02-2025T14:00:00" --days 7 > opsgenie-report.md
"""

import io
import json
import logging
import re
from contextlib import redirect_stdout
from datetime import datetime, timedelta, timezone

import click
from aika import TimeInterval, TimeIntervalParser
from munch import munchify
from opsgenie_sdk import AlertApi, ApiClient, ApiException, Configuration
from opsgenie_sdk.exceptions import ConfigurationException
from tabulate import tabulate

logger = logging.getLogger(__name__)


class OpsgenieAlertsClient:
    """
    Fetch alert items from Opsgenie API.
    """

    OPSGENIE_DATETIME_FORMAT = "%d-%m-%YT%H:%M:%S"

    def __init__(self, api_key: str, query: str):
        self.api_key = api_key
        self.query = query
        self.api: ApiClient = self.client_factory(self.api_key)

    @classmethod
    def format_interval(cls, interval: TimeInterval):
        """Format the interval for the query expression."""
        expression = f'createdAt >= "{interval.start.strftime(cls.OPSGENIE_DATETIME_FORMAT)}"'
        if interval.end:
            expression += (
                f' and createdAt <= "{interval.end.strftime(cls.OPSGENIE_DATETIME_FORMAT)}"'
            )
        return expression

    @classmethod
    def query_from_cli_options(cls, ctx: click.Context) -> str:
        """Build the Opsgenie query string based on CLI arguments."""
        params = munchify(ctx.params)
        if params.when:
            tr = TimeIntervalParser()
            interval = tr.parse(params.when)
        elif params.start_time:
            if params.days is None:
                raise ValueError("--days is required when --start-time is provided")
            start_dt = datetime.strptime(params.start_time, cls.OPSGENIE_DATETIME_FORMAT)
            start_dt = start_dt.replace(tzinfo=timezone.utc)
            end_dt = start_dt + timedelta(days=params.days)
            interval = TimeInterval(start_dt, end_dt)
        else:
            seven_days_ago = datetime.now(timezone.utc) - timedelta(days=7)
            interval = TimeInterval(seven_days_ago, None)
        query = cls.format_interval(interval)
        logger.info(f"Using query: {query}")
        return query

    @staticmethod
    def client_factory(api_key: str) -> ApiClient:
        """Create an Opsgenie ApiClient instance."""
        configuration = Configuration()
        configuration.api_key["Authorization"] = api_key
        return ApiClient(configuration)

    def fetch(self, limit: int = 100) -> list:
        """Fetch all alerts from Opsgenie that match the given query, paginating as needed."""
        alerts = []
        offset = 0
        alert_api = AlertApi(api_client=self.api)
        while True:
            response = alert_api.list_alerts(
                query=self.query, limit=limit, offset=offset, sort="createdAt", order="asc"
            )
            if not response.data:
                break
            alerts.extend(response.data)
            offset += limit
            if len(response.data) < limit:
                break
        return alerts


class OpsgenieAlertsReport:
    """
    Report about alert items from Opsgenie.
    """

    def __init__(self, client: OpsgenieAlertsClient) -> None:
        self.client: OpsgenieAlertsClient = client
        self.alerts_data: list
        self.summary_data: list
        self.token_summary_data: list

    def process(self) -> None:
        """
        Process alerts into data tables.

        Returns:
            alerts_data: list of [Trigger Date, Priority, Alert Message, Resolved Time, Open Duration]
            summary_data: list of [Priority, Total Count, Out-of-hours Count]
            token_summary_data: list of [First Alert Token, Count]
        """  # noqa: E501
        alerts_data: list = []
        priority_count: dict = {}
        out_of_hours_count: dict = {}
        message_token_count: dict = {}

        # Patterns to remove certain strings from the alert message.
        remove_patterns = [
            r"\[Prometheus\]: \[FIRING:.*\] ",
            r"kubernetes-service-endpoints",
            r"crate",
            r"prod",
            r"kubernetes-nodes",
            r"stable",
            r"db cloud",
            r"( metrics p)",
            r"\(metrics promethe",
        ]

        try:
            alerts = self.client.fetch()
        except ConfigurationException as e:
            msg = f"Opsgenie configuration error: {e.reason}; {json.loads(e.body)['message']}"
            logger.error(msg)
            raise ValueError(msg) from e
        except ApiException as e:
            msg = f"Opsgenie API error: {e.reason}"
            logger.error(msg)
            raise IOError(msg) from e

        for alert in alerts:
            created_at_str = alert.created_at.strftime("%Y-%m-%d %H:%M")
            raw_message = alert.message
            for pattern in remove_patterns:
                raw_message = re.sub(pattern, "", raw_message)
            message = raw_message.strip()  # Adjust truncation here if necessary

            # Count first token from message
            tokens = message.split()
            first_token = tokens[0] if tokens else ""
            message_token_count[first_token] = message_token_count.get(first_token, 0) + 1

            # Determine resolved values and open duration
            if alert.report.close_time:
                resolved_dt = alert.created_at + timedelta(milliseconds=alert.report.close_time)
                resolved_time = resolved_dt.strftime("%Y-%m-%d %H:%M")
                open_duration_str = str(resolved_dt - alert.created_at).split(".")[0]
            else:
                resolved_time = "Open"
                open_duration_str = "-"

            priority = alert.priority if alert.priority else "Not Specified"
            priority_count[priority] = priority_count.get(priority, 0) + 1

            if alert.created_at.hour >= 22 or alert.created_at.hour < 8:
                out_of_hours_count[priority] = out_of_hours_count.get(priority, 0) + 1

            alerts_data.append(
                [created_at_str, priority, message, resolved_time, open_duration_str]
            )

        # Prepare summary data for priorities and out-of-hours count.
        summary_data = []
        for prio, total in priority_count.items():
            outside = out_of_hours_count.get(prio, 0)
            summary_data.append([prio, total, outside])

        # Prepare summary data for the first token occurrences.
        token_summary_data = []
        for token, count in sorted(message_token_count.items()):
            token_summary_data.append([token, count])

        # Sort alerts by creation time.
        alerts_data.sort(key=lambda row: row[0])

        self.alerts_data = alerts_data
        self.summary_data = summary_data
        self.token_summary_data = token_summary_data

    def to_markdown(self) -> str:
        """Generate Markdown output."""
        md = (
            f"## Opsgenie Alerts Report ({self.client.query})\n\n"
            "| Trigger Date       | Prty           | Alert Message                              | Resolved Time    | Open Duration |\n"  # noqa: E501
            "|--------------------|----------------|--------------------------------------------|------------------|---------------|\n"
        )
        md += "\n".join(
            [
                f"| {row[0]} | {row[1]} | {row[2]} | {row[3]} | {row[4]} |"
                for row in self.alerts_data
            ]
        )
        md += "\n\n## Alert Type Summary\n\n"
        md += (
            "| Priority Level     | Count | Out-of-hours |\n"
            "|--------------------|-------|--------------|\n"
        )
        md += "\n".join([f"| {row[0]} | {row[1]} | {row[2]} |" for row in self.summary_data])
        md += "\n\n## Alert Message Summary\n\n"
        md += "| Alert Token       | Count |\n|-------------------|-------|\n"
        md += "\n".join([f"| {row[0]} | {row[1]} |" for row in self.token_summary_data])
        return md

    def to_text(self) -> str:
        """Generate formatted terminal output."""
        with redirect_stdout(io.StringIO()) as stdout:
            print("Formatted Terminal Output:\n")
            print(
                tabulate(
                    self.alerts_data,
                    headers=[
                        "Trigger Date",
                        "Priority",
                        "Alert Message",
                        "Resolved Time",
                        "Open Duration",
                    ],
                    tablefmt="grid",
                )
            )
            print("\nAlert Type Summary:")
            print(
                tabulate(
                    self.summary_data,
                    headers=["Priority Level", "Count", "Out-of-hours"],
                    tablefmt="grid",
                )
            )
            print("\nAlert Message Summary:")
            print(
                tabulate(self.token_summary_data, headers=["Alert Token", "Count"], tablefmt="grid")
            )
        stdout.seek(0)
        return stdout.read()



"""
Use Slack conversations like a static documentation generator target.
"""

import dataclasses
import datetime as dt
import logging
import typing as t

from pueblo_goof.slack.conversation import SlackConversation
from rapporto import __version__
from rapporto.report.model import DailyItem, ReportOptions, WeeklyReport
from rapporto.source.github.model import GitHubOptions

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class SlackWeekly:
    """
    Publish daily reports to Slack, in a per-week context, with hourly granularity.
    """

    ROOT_EVENT = "qabot_root_created"
    PREAMBLE_EVENT = "qabot_preamble_created"
    ITEM_EVENT = "qabot_item_created"
    AUTHOR = "qa-bot"

    week: str
    github_options: GitHubOptions
    report_options: ReportOptions
    conversation: SlackConversation

    # Message id of the root message.
    root_id: t.Optional[str] = dataclasses.field(default=None)

    def __post_init__(self):
        if self.week is None:
            self.week = dt.datetime.now().strftime("%YW%V")

    def refresh(self):
        self.seed()
        self.render()

    def seed(self, just_created: bool = False):
        """
        Create/update and seed the conversation.
        """

        logger.info(f"Creating or updating conversation for calendar week: {self.week}")

        message = self.conversation.find_message_by_metadata(
            self.conversation.messages(), type="root", week=self.week
        )
        metadata = {
            "author": self.AUTHOR,
            "type": "root",
            "week": self.week,
        }
        if message:
            self.root_id = message["ts"]
            if self.root_id is None:
                raise KeyError("Root message was not created")
            self.conversation.update(
                ts=self.root_id, markdown=self.root_markdown, metadata=metadata
            )
        else:
            # Submit the root message.
            response = self.conversation.send(
                markdown=self.root_markdown, event=self.ROOT_EVENT, metadata=metadata
            )
            self.root_id = response["ts"]
            if self.root_id is None:
                raise KeyError("Root message was not created")

        # Submit or update the preamble message.
        preamble = DailyItem(type="preamble", day=self.week, markdown=self.preamble_markdown)
        self.create_or_update_item(item=preamble, msg_type="preamble", event=self.PREAMBLE_EVENT)

    @property
    def root_markdown(self):
        """
        The message body for the root message, in Markdown format.
        """
        return f"# qa-bot {self.week}"

    @property
    def preamble_markdown(self):
        """
        The message body for the preamble message, in Markdown format.
        """
        timestamp = dt.datetime.now().replace(microsecond=0).isoformat()
        conversation_link = (
            self.root_id and f"[🔗]({self.conversation.get_permalink(self.root_id)})"
        ) or ""
        changelog_link = "[🔗](https://rapporto.readthedocs.io/changes.html)"
        items = [
            f"**Week:** {self.week}",
            f"**Updated:** {timestamp}",
            f"**Root message:** {self.root_id}  {conversation_link}",
            f"**Producer:** Rapporto v{__version__}  {changelog_link}",
        ]
        return "\n".join(items)

    def render(self):
        weekly = WeeklyReport(
            week=self.week,
            github_options=self.github_options,
            report_options=self.report_options,
        )
        weekly.process()
        for daily_report in weekly.dailies:
            for daily_item in daily_report.items:
                self.create_or_update_item(daily_item)

    def create_or_update_item(
        self, item: DailyItem, msg_type: str = "item", event: str = ITEM_EVENT
    ):
        """
        Create or update a Slack message representing a DailyItem.

        TODO: Add timestamp fields to metadata, and also update them on Slack's `update` operations.
        TODO: Generalize to also use with conversation's `seed` operation.
        """
        key = f"{item.type}_{item.day}"

        if self.root_id is None:
            raise KeyError("Unable to create items without root message")
        logger.info(f"Creating or updating reply for key: {key}")
        message = self.conversation.find_message_by_metadata(
            self.conversation.replies(ts=self.root_id), type=msg_type, key=key
        )
        metadata = {
            "key": key,
            "type": msg_type,
            "author": self.AUTHOR,
            "week": self.week,
            "day": item.day,
        }
        if message:
            message_id = message["ts"]
            logger.info(f"Updating message with key: {key}")
            return self.conversation.update(
                ts=message_id, markdown=item.markdown, metadata=metadata
            )
        else:
            logger.info(f"Sending message with key: {key}")
            return self.conversation.send(
                markdown=item.markdown,
                reply_to=self.root_id,
                event=event,
                metadata=metadata,
            )



import click

from rapporto.source.opsgenie.core import OpsgenieAlertsClient, OpsgenieAlertsReport


@click.group()
@click.pass_context
def cli(ctx: click.Context):
    """
    Harvest information from Opsgenie.
    """
    pass


@cli.command()
@click.option(
    "--api-key", type=str, envvar="OPSGENIE_API_KEY", required=False, help="Opsgenie API key"
)
@click.option("--when", type=str, required=False, help="A humanized time interval expression")
@click.option(
    "--start-time", type=str, required=False, help="Start time in format dd-mm-yyyyTHH:MM:SS"
)
@click.option("--days", type=int, required=False, help="Number of days from start time")
@click.option(
    "--format",
    "format_",
    type=click.Choice(["md", "text"]),
    required=False,
    default="md",
    help="Output format: 'md' for markdown, 'text' for nicely formatted terminal output",
)
@click.pass_context
def export_alerts(
    ctx: click.Context, api_key: str, when: str, start_time: str, days: int, format_: str
):
    """
    Report about alerts in Opsgenie.
    """
    if not api_key:
        raise click.UsageError(
            "Missing option '--api-key' or environment variable 'OPSGENIE_API_KEY'."
        )

    client = OpsgenieAlertsClient(
        api_key=api_key, query=OpsgenieAlertsClient.query_from_cli_options(ctx)
    )
    report = OpsgenieAlertsReport(client=client)

    try:
        report.process()
    except (ValueError, IOError) as e:
        raise SystemExit(1) from e

    if format_ == "md":
        print(report.to_markdown())
    else:
        print(report.to_text())



import logging
import sys
import typing as t
from pathlib import Path

import click
from slack_sdk.errors import SlackApiError

from pueblo_goof.slack.conversation import SlackConversation
from pueblo_goof.slack.model import slack_api_token_option
from pueblo_goof.util import Zapper

logger = logging.getLogger(__name__)


@click.group()
@click.pass_context
def cli(ctx: click.Context):
    """
    Goofing. Yeah.
    """
    pass


@cli.group()
@slack_api_token_option
@click.pass_context
def slack(ctx: click.Context, slack_token: str):
    """
    Send information to Slack.
    """
    if not slack_token:
        raise click.UsageError(
            "Missing option '--slack-token' or environment variable 'SLACK_TOKEN'."
        )
    ctx.meta.update({"slack_token": slack_token})


@slack.command()
@click.option(
    "--message",
    "-m",
    type=str,
    multiple=True,
    required=True,
    help="Text of the message in Markdown format",
)
@click.option("--channel", "-c", type=str, required=False, help="Target channel (name, ID, or URL)")
@click.option("--reply-to", "-r", type=str, required=False, help="Message to reply to (ID or URL)")
@click.option("--update", "-u", type=str, required=False, help="Message to update (ID or URL)")
@click.option("--zap", "-z", type=str, required=False, help="Zap message again")
@click.pass_context
def send(
    ctx: click.Context, channel: str, reply_to: str, update: str, message: t.List[str], zap: str
) -> None:
    """
    Send one or multiple Slack messages in Markdown format.
    """

    if not (channel or update or reply_to):
        raise click.UsageError("Please provide either 'channel' or 'update' or 'reply_to'")

    channel = channel or update or reply_to

    conversation = SlackConversation(api_token=ctx.meta["slack_token"], channel=channel)
    zapper = Zapper(when=zap, action=conversation.delete)

    message_ids = []
    for msg in message:
        if msg == "-":
            msg = sys.stdin.read()
        elif Path(msg).exists():
            msg = Path(msg).read_text()
        if update:
            response = conversation.update(ts=update, markdown=msg)
        else:
            response = conversation.send(markdown=msg, reply_to=reply_to)
        message_id = response["ts"]
        message_ids.append(message_id)

    try:
        zapper.process()
    except Exception as e:
        raise click.ClickException(f"The 'zap' operation failed: {e}") from e


@slack.command()
@click.option("--channel", "-c", type=str, required=False, help="Target channel (name, ID, or URL)")
@click.option(
    "--id",
    "identifiers",
    type=str,
    multiple=True,
    required=True,
    help="Message to delete (ID or URL)",
)
@click.pass_context
def delete(ctx: click.Context, channel: str, identifiers: t.List[str]):
    """
    Delete individual Slack message by ID or URL, also multiple ones.
    """
    if not (channel or identifiers):
        raise click.UsageError("Please provide either 'channel' and/or 'identifiers'")

    # TODO: Currently, when no channel is provided, but fully-qualified message identifiers,
    #       conveyed through Slack URLs, the channel is derived from the first identifier.
    #       To make it work in all situations, this code will need to iterate all identifiers,
    #       and consider each one individually, as it might concern different channels.
    channel = channel or identifiers[0]

    conversation = SlackConversation(api_token=ctx.meta["slack_token"], channel=channel)
    had_errors = False
    for identifier in identifiers:
        try:
            conversation.delete_message(identifier=identifier)
        except SlackApiError as ex:
            msg = f"Deleting message failed: {ex}"
            logger.error(msg)
            had_errors = True
    if had_errors:
        raise click.ClickException(
            "The 'delete' operation failed or was not completely successful."
        )



import dataclasses
import logging
import typing as t
from collections import OrderedDict

from aika import TimeInterval
from munch import Munch, munchify
from tqdm import tqdm

from rapporto.source.github.model import (
    GitHubMultiRepositoryInquiry,
    MarkdownContent,
    timeinterval,
)
from rapporto.source.github.util import GitHubHttpClient
from rapporto.util import sanitize_title

logger = logging.getLogger(__name__)


class GitHubActionsReport:
    """
    Report about failed outcomes of GitHub Actions workflow runs.
    """

    def __init__(self, inquiry: GitHubMultiRepositoryInquiry):
        self.inquiry = inquiry
        self.request = GitHubActionsRequest(inquiry)
        self.runs_failed = self.request.runs_failed
        self.runs_pr_success = self.request.runs_pr_success

    @property
    def runs(self):
        """
        All failed runs, modulo duplicates and subsequent succeeding PR runs.
        """
        seen = set()
        for run in self.runs_failed:
            # Filter duplicates.
            key = f"{run.repository.full_name}-{run.head_branch}"
            if key in seen:
                continue

            # Filter PRs that subsequently succeeded.
            if run.event == "pull_request" and self.is_pr_successful(run):
                continue

            seen.add(key)
            yield run

    def is_pr_successful(self, run):
        """
        Find out if a given run has others that succeeded afterward.
        """
        for pr in self.runs_pr_success:
            if (
                run.repository.full_name == pr.repository.full_name
                and run.head_branch == pr.head_branch
            ):
                if pr.conclusion == "success":
                    return True
        return False

    @property
    def markdown(self):
        mdc = MarkdownContent(labels=self.request.event_section_map)
        for run in self.runs:
            mdc.add(run.event, run.markdown)
        return f"""
# CI failures report {self.request.timeinterval.start.date().isoformat()}

A report about GitHub Actions workflow runs that failed recently.
Time range: {self.request.timeinterval.githubformat() or "n/a"}
{mdc.render()}
        """.strip()


class GitHubActionsRequest:
    """
    Fetch outcomes of recent GitHub Actions workflow runs.

    Possible event types are: dynamic, pull_request, push, schedule
    """

    event_section_map: t.ClassVar[t.OrderedDict[str, str]] = OrderedDict(
        schedule="Schedule",
        pull_request="Pull requests",
        # push="Pushes",
        dynamic="Dynamic",
    )

    def __init__(self, inquiry: GitHubMultiRepositoryInquiry):
        self.inquiry = inquiry
        self.session = GitHubHttpClient.session

    @property
    def timeinterval(self) -> TimeInterval:
        """
        Return same-day time interval.

        TODO: Expand to use other, more broad time intervals sensibly.
        """
        ti = timeinterval(self.inquiry.created)
        ti.end = ti.start
        return ti

    @property
    def created(self) -> str:
        """
        Compute the start timestamp in ISO format.
        Truncate the ISO format after the hour, to permit caching.
        """
        # TODO: What about `%Y-%m-%dT%H`?
        # TODO: What about `dt.timedelta(hours=self.DELTA_HOURS)`, with `DELTA_HOURS = 24`?
        return self.timeinterval.githubformat()

    def fetch(self, filter: "ActionsFilter") -> t.List["ActionsOutcome"]:  # noqa:A002
        outcomes = []
        for repository in tqdm(
            self.inquiry.repositories,
            desc=f"Fetching failed GitHub Actions outcomes for event={filter.event}",
            leave=False,
        ):
            url = f"https://api.github.com/repos/{repository}/actions/runs?{filter.query}"
            logger.debug(f"Using API URL: {url}")
            response = self.session.get(url)
            if response.status_code == 404:
                continue
            response.raise_for_status()
            for run in munchify(response.json()).workflow_runs:
                outcome = ActionsOutcome(
                    id=run.id,
                    event=run.event,
                    status=run.status,
                    conclusion=run.conclusion,
                    repository=run.repository,
                    name=run.display_title,
                    url=run.html_url,
                    started=run.run_started_at,
                    head_branch=run.head_branch,
                )
                outcomes.append(outcome)
        return outcomes

    @property
    def runs_failed(self):
        return self.fetch(filter=ActionsFilter(status="failure", created=self.created))

    @property
    def runs_pr_success(self):
        return self.fetch(
            filter=ActionsFilter(event="pull_request", status="success", created=self.created)
        )


@dataclasses.dataclass
class ActionsFilter:
    event: t.Optional[str] = None
    status: t.Optional[str] = None
    created: t.Optional[str] = None

    @property
    def query(self) -> str:
        expression = []
        if self.event:
            expression.append(f"event={self.event}")
        if self.status:
            expression.append(f"status={self.status}")
        if self.created:
            expression.append(f"created={self.created}")
        return "&".join(expression)


@dataclasses.dataclass
class ActionsOutcome:
    id: int
    event: str
    status: str
    conclusion: str
    repository: Munch
    name: str
    url: str
    started: str
    head_branch: str

    @property
    def markdown(self):
        title = sanitize_title(f"{self.repository.full_name}: {self.name}")
        return f"- [{title}]({self.url})"



import dataclasses
import logging
import typing as t
from abc import abstractmethod
from operator import attrgetter
from textwrap import dedent

from attrs import define
from dataclasses_json import CatchAll, Undefined, dataclass_json
from requests import Session
from tqdm import tqdm

from rapporto.source.github.model import (
    GitHubInquiry,
    GitHubQueryBuilder,
    GitHubSearch,
)
from rapporto.source.github.util import GitHubHttpClient, repository_name
from rapporto.util import sanitize_title

logger = logging.getLogger(__name__)


class GitHubActivityQueryBuilder(GitHubQueryBuilder):
    """
    Find all open issues and pull requests created by individual author.
    """

    def query(self):
        self.add("org", self.inquiry.organization, is_list=True)
        self.add("author", self.inquiry.author, is_list=True)
        self.add("updated", self.timeinterval.githubformat())


class GitHubActivityReport:
    """
    Report about activity across a whole GitHub organization.
    """

    def __init__(self, inquiry: GitHubInquiry):
        self.inquiry = inquiry
        self.session = GitHubHttpClient.session
        self.search = GitHubSearch.with_query_builder(
            self.session, GitHubActivityQueryBuilder(inquiry=inquiry)
        )

    @property
    def repository_names(self):
        items = self.search.issues_and_prs()
        names = []
        for item in items:
            names.append(repository_name(item["repository_url"]))
        return sorted(set(names))

    @property
    def markdown_overview(self):
        link_issues = f"[Issues]({self.search.issues_html})"
        link_pulls = f"[Pull requests]({self.search.pulls_html})"
        return dedent(f"""
        ## Overview
        *Progress:*
          - About: Bugfixes, Documentation, Guidance, Planning, Support
          - Activity: {", ".join(self.repository_names)}
          - Details: {link_issues}, {link_pulls}
        *Plans:* Dito.
        *Problems:* n/a
        """).strip()

    @property
    def markdown_significant_issues(self):
        return GitHubSignificantIssues(session=self.session, search=self.search).markdown

    @property
    def markdown_significant_prs(self):
        return GitHubSignificantPullRequests(session=self.session, search=self.search).markdown

    @staticmethod
    def format_pr(item: "PullRequestMetadata"):
        return (
            f"comments: {item.comments_total}, files: {item.changed_files}, size: {item.code_size}"
        )

    @property
    def markdown(self) -> str:
        timerange = (self.inquiry.updated and f"for {self.inquiry.updated}") or ""
        return f"""
# Activity report {timerange}

{self.markdown_overview}

## Top issues
{self.markdown_significant_issues}

## Top changes
{self.markdown_significant_prs}
""".strip()


@dataclass_json(undefined=Undefined.INCLUDE)
@dataclasses.dataclass
class IssueMetadata:
    """
    Manage attributes for a GitHub API issue item.
    """

    # Original fields from GitHub API.
    # TODO: Also count reactions.
    number: str
    url: str
    html_url: str
    title: str
    comments: int

    # Computed fields.
    comments_total: t.Optional[int] = None

    # This dictionary includes all the remaining fields.
    more: t.Optional[CatchAll] = None

    def __post_init__(self):
        self.comments_total = self.comments  # + self.review_comments
        try:
            self.repo_name = repository_name(self.html_url, with_org=True)
        except KeyError as ex:
            msg = f"Unable to decode repository name: {ex}"
            logger.error(msg)
            raise KeyError(msg) from ex

    def format_item(self):
        return f"  - [{self.repo_name}: {sanitize_title(self.title)}]({self.html_url})"


@dataclass_json(undefined=Undefined.INCLUDE)
@dataclasses.dataclass
class PullRequestMetadata:
    """
    Manage attributes for a GitHub API pull request item.
    """

    # Original fields from GitHub API.
    number: str
    url: str
    html_url: str
    title: str
    commits: int
    additions: int
    deletions: int
    changed_files: int
    comments: int
    review_comments: int

    # Computed fields.
    code_size: t.Optional[int] = None
    comments_total: t.Optional[int] = None
    repo_name: t.Optional[str] = None

    # This dictionary includes all the remaining fields.
    more: t.Optional[CatchAll] = None

    def __post_init__(self):
        self.code_size = self.additions - self.deletions
        self.comments_total = self.comments + self.review_comments
        try:
            self.repo_name = self.more["base"]["repo"]["name"]  # type: ignore[index]
        except KeyError as ex:
            msg = f"Unable to decode repository name: {ex}"
            logger.error(msg)
            raise KeyError(msg) from ex

    def format_item(self):
        # Generic info (for debugging)
        # return (f"files: {self.changed_files}, "
        #        f"size: {self.code_size}, comments: {self.comments_total}")
        # Repo: PR+link
        return f"  - [{self.repo_name}: {sanitize_title(self.title)}]({self.html_url})"


@define
class SignificantItemsBase:
    """
    Common base class for significant items.
    """

    session: Session
    search: GitHubSearch
    metadata_class: t.ClassVar[t.Type[t.Union[IssueMetadata, PullRequestMetadata]]]
    description: t.ClassVar[str]
    by_size_sort_attributes: t.ClassVar[t.List[str]]
    by_comments_sort_attributes: t.ClassVar[t.List[str]]

    @property
    @abstractmethod
    def api_url(self):
        """
        Implemented individually between Issue vs. PRs.
        """
        raise NotImplementedError()

    @staticmethod
    @abstractmethod
    def decode_url(item):
        """
        Implemented individually between Issue vs. PRs.
        """
        raise NotImplementedError()

    def items(self):
        """
        Acquire items from GitHub API.
        """
        items = []
        response = self.session.get(self.api_url)
        response.raise_for_status()
        for item in tqdm(response.json()["items"], desc=self.description, leave=False):
            url = self.decode_url(item)
            response = self.session.get(url)
            response.raise_for_status()
            data = response.json()
            cm = self.metadata_class.from_dict(data)  # type: ignore[attr-defined,union-attr]
            items.append(cm)
        return items

    def significant(self):
        """
        Return 2/5 of the most significant PRs, or any other share.

        The list of candidates (all PRs within given time range) is sorted by
        number of comments, number of changed files, and delta code size.
        """
        items_in = self.items()
        items_out = []

        items_max = int(len(items_in) / 2 / 5) + 1

        by_size_items = sorted(
            items_in, key=attrgetter(*self.by_size_sort_attributes), reverse=True
        )
        items_out += by_size_items[:items_max]

        by_comments_items = sorted(
            items_in, key=attrgetter(*self.by_comments_sort_attributes), reverse=True
        )
        # TODO: Need to deduplicate manually.
        # items += by_comments_items[:items_max]
        for item in by_comments_items[:items_max]:
            if item not in items_out:
                items_out.append(item)

        return items_out

    @property
    def markdown(self):
        items = [item.format_item() for item in self.significant()]
        return "\n".join(items)


@define
class GitHubSignificantIssues(SignificantItemsBase):
    """
    Discover significant issues from GitHub API.
    """

    metadata_class = IssueMetadata
    description = "Fetching issues"
    by_size_sort_attributes: t.ClassVar[t.List[str]] = ["comments_total"]
    by_comments_sort_attributes: t.ClassVar[t.List[str]] = ["comments_total"]

    @property
    def api_url(self):
        return self.search.issues_api

    @staticmethod
    def decode_url(item):
        return item["url"]


@define
class GitHubSignificantPullRequests(SignificantItemsBase):
    """
    Discover significant pull requests from GitHub API.
    """

    metadata_class = PullRequestMetadata
    description = "Fetching pull requests"
    by_size_sort_attributes: t.ClassVar[t.List[str]] = [
        "code_size",
        "changed_files",
        "comments_total",
    ]
    by_comments_sort_attributes: t.ClassVar[t.List[str]] = [
        "comments_total",
        "changed_files",
        "code_size",
    ]

    @property
    def api_url(self):
        return self.search.pulls_api

    @staticmethod
    def decode_url(item):
        return item["pull_request"]["url"]



"""
Aggregated, chronological change log across multiple projects.

References:
- zt.manticore.ext.changes
"""

import copy
import glob
import logging
import os
import re
import sys
import typing as t
from collections import namedtuple
from operator import attrgetter
from pathlib import Path

from .util import mkproject, rest_header, walk_projects

HACKS: t.Dict[str, t.Dict[str, str]] = {
    "project_aliases": {
        # 'from': 'to',
    }
}

Segment = namedtuple("Segment", "date, version, author, text")
Change = namedtuple("Change", "date, name, version, author, text")


logger = logging.getLogger(__name__)


class ChangesFileReader:
    def __init__(self, changes_file_path):
        self.changes_file_path = changes_file_path

    def expand_roles(self, line):
        """
        Expand strings like 'BUG-XXXX' to ':bug:`BUG-XXXX`'.
        For using with the `sphinx.ext.extlinks` extension.
        """
        expansions = [("BUG", ":bug:")]
        for expansion in expansions:
            symbol = expansion[0]
            role = expansion[1]
            if (
                "{symbol}-".format(**locals()) in line
                and "{role}`{symbol}-".format(**locals()) not in line
            ):
                line = re.sub(
                    r"{symbol}-(\d+)".format(**locals()),
                    r"{role}`{symbol}-\1`".format(**locals()),
                    line,
                )
        return line

    def lines(self):
        for line in Path(self.changes_file_path).read_text().splitlines():
            yield self.expand_roles(line)


class ChangesFileSegmentizer:
    """CHANGES file parser and segmentizer. Uses regexes and a little state machine."""

    def __init__(self, reader):
        self.reader = reader
        self.date_patterns = [
            r"(?P<year>\d{4})[/-](?P<month>\d{2})[/-](?P<day>\d{2})?",
            r"(?P<day>\d{2})[/-](?P<month>\d{2})[/-](?P<year>\d{4})?",
        ]
        self.header_patterns = []
        self.suffix_pattern = r"""
            (?:\s?\(?v?
                (?P<version>[-.0-9]+)
            \)?)?
            (?:\s\[(?P<author>\D+?)\])?
            :?
        """
        for date_pattern in self.date_patterns:
            pattern = date_pattern + self.suffix_pattern
            # print pattern
            self.header_patterns.append(re.compile(pattern, re.VERBOSE))
        self.date_tpl = "%(year)s-%(month)s-%(day)s"
        self.date_tpl = "%(version)s"

    def _match_header(self, line):
        for header_pattern in self.header_patterns:
            m = header_pattern.match(line)
            if m:
                return m
        return None

    def get_entries(self):
        state_header = False
        state_inblock = False

        # initialize data variables
        date = None
        version = None
        author = None
        block: t.List[str] = []

        def response():
            # yield out response entry
            segment = Segment(date, version, author, "\n".join(block))
            # print segment
            return segment

        for line in self.reader.lines():
            # if not line: continue
            line = line.rstrip()

            # check line for "date & version" pattern
            m = self._match_header(line)
            if m:
                if state_inblock:
                    # yield out finished block
                    yield response()

                    # reset for next iteration
                    date = None
                    version = None
                    author = None
                    block = []

                # compute normalized change entry attributes
                payload = m.groupdict()
                date = "%(year)s-%(month)s-%(day)s" % payload
                version = payload["version"]
                author = payload["author"]
                state_header = True
                continue

            if state_header:
                # skip reST title underlines
                if line.startswith("=" * 5) or line.startswith("-" * 5):
                    continue

                # switch state to block content aggregation
                else:
                    state_header = False
                    state_inblock = True

            # aggregate block content
            if state_inblock:
                block.append(line)

        # yield out last block
        yield response()


class ChangesAggregator:
    """Aggregates and holds all global changes. Provides a sorted activity stream of changes."""

    def __init__(self, project_path, summary_path):
        # configuration data
        self.filename_choices = [
            "CHANGES.rst",
            "CHANGELOG.rst",
            "CHANGES.txt",
            "CHANGELOG.txt",
            "CHANGES",
            "HISTORY",
        ]
        self.project_path = project_path
        self.summary_path = summary_path

        # volatile data
        self.changes = set()
        self.projects = set()

    def walk_projects_changes_files(self):
        """Scans all main level directories in the projects directory for a 'CHANGES.rst' file"""
        p = mkproject(self.project_path)
        if p is None:
            projects = walk_projects(self.project_path)
        else:
            projects = [p]
        for project in projects:
            possible_files = glob.glob(project.path + "/*") + glob.glob(project.path + "/*/*")
            changes_files_found = [
                changes_file
                for changes_file in possible_files
                if os.path.basename(changes_file) in self.filename_choices
            ]
            if not changes_files_found:
                yield project
            for changes_file_found in changes_files_found:
                project_ready = copy.deepcopy(project)
                project_ready.changes_file = changes_file_found
                yield project_ready

    def normalize_project_name(self, project_name):
        for alias, name in HACKS.get("project_aliases", {}).items():
            project_name = project_name.replace(alias, name)
        return project_name

    def read_changes_file(self, project):
        """Read 'CHANGES.rst', segmentize into change entries and store in memory"""
        # print changes_file_path
        # project_name = self.normalize_project_name(project.name)
        if not project.changes_file:
            return
        cfs = ChangesFileSegmentizer(ChangesFileReader(project.changes_file))
        for entry in cfs.get_entries():
            entry = Change(entry.date, project.name, entry.version, entry.author, entry.text)
            self.changes.add(entry)

    def compute_changes(self):
        # scan for CHANGES files
        for project in self.walk_projects_changes_files():
            project.name = self.normalize_project_name(project.name)
            self.projects.add(project)
            self.read_changes_file(project)

        logger.info(f"Processing projects: {self.projects}")

        # process change entries
        self.changes = sorted(self.changes, key=attrgetter("date", "name", "version"))  # type: ignore[assignment]

    def get_change_title(self, change, short=False):
        if short:
            change_name = change.name or ""
            change_version = change.version or ""
            return change_name + " " + change_version
        else:
            change_header = "%(date)s :ref:`%(name)s <%(name)s>`" % change._asdict()
            if change.version:
                change_header += " " + change.version
            if change.author:
                change_header += " [author: " + change.author + "]"
            return change_header

    def get_change_message(self, change):
        message = ""
        if change.date is None or change.date == "":
            message += ".. note:: Date missing\n"
        # if change.version is None or change.version == '':
        #    message += '.. note:: Version missing'
        if change.text is None or change.text == "":
            message += ".. note:: Change text is empty"
        return message

    def write_summary_rst(self):
        summary_file = os.path.join(self.summary_path, "changes.rst")
        summary_path = os.path.dirname(summary_file)
        if not os.path.exists(summary_path):
            os.makedirs(summary_path)
        project_names = [project.name for project in self.projects]
        project_names.sort()
        # f = open(summary_file, "w")
        f = sys.stdout
        f.write(rest_header("Release notes", "rapporto"))
        f.write(
            "Aggregated release notes across multiple projects' change log files\n"
            "in reverse chronological order, to be read as an activity stream.\n\n"
        )
        f.write("Project names: ")
        f.write(", ".join(project_names))
        # f.write("\n\n")
        # f.write(self.get_timeline_widget())
        f.write("\n\n\n")

        changes = copy.copy(self.changes)
        changes = reversed(list(changes))  # type: ignore[assignment]
        for change in changes:
            title = self.get_change_title(change)
            f.write(title + "\n")
            f.write("-" * len(title) + "\n")
            f.write(self.sanitize_change_text(change.text) + "\n")
            f.write(self.get_change_message(change))
            f.write("\n\n")

        f.close()

    def get_timeline_widget(self):
        return ".. seealso:: :doc:`timeline`"
        timeline_widget = """
----

Timeline data: :download:`changes.js`.

.. raw:: html

    <!-- Include Simile Timeline Widget -->
    <iframe src="../_static/timeline/timeline.html" width="100%" height="400" style="border: 0px; overflow: visible">
        <p>Changes timeline</p>
    </iframe>

----
        """  # noqa: E501
        return timeline_widget

    def write_summary_js(self):
        summary = """var timeline_data = {
    'dateTimeFormat': 'iso8601',
    'wikiURL': "http://manticore.example.net/summary/changes.html",
    'wikiSection': "project changes",
    'events': [
        %(events)s
    ]
}
        """

        """
        'events': [
            {'start': '2009-12-13',
            'title': 'Barfusserkirche',
            'description': 'by Lyonel Feininger, American/German Painter, 1871-1956',
            'image': 'http://images.allposters.com/images/AWI/NR096_b.jpg',
            'link': 'http://www.allposters.com/-sp/Barfusserkirche-1924-Posters_i1116895_.htm'
            },
        ]
        """
        event_template = """
        {
            'start': '%(start_date)s',
            'title': '%(title)s',
            'description': '%(description)s',
            'image': '%(image)s',
            'link': '%(link)s'
        },
        """

        def sanitize_js(text):
            return text.replace("'", '"').replace("\n-", "<br/><br/>-").replace("\n", "<br/>")

        events = []
        changes = copy.copy(self.changes)

        for change in changes:
            start_date = change.date
            if not start_date:
                continue
            title = sanitize_js(self.get_change_title(change, short=True))
            description = sanitize_js(self.sanitize_change_text(change.text))
            image = ""  # TODO
            link = ""  # TODO
            description += sanitize_js(self.get_change_message(change))
            # print locals()
            events.append(event_template % locals())

        summary_file = os.path.join(self.summary_path, "changes.js")
        summary_path = os.path.dirname(summary_file)
        if not os.path.exists(summary_path):
            os.makedirs(summary_path)
        f = open(summary_file, "w")
        f.write(summary % {"events": "".join(events)})
        f.close()

    def sanitize_change_text(self, text):
        lines = []
        for line in text.split("\n"):
            if re.match("^[-=]+$", line.strip()):
                continue
            lines.append(line)
        return "\n".join(lines)

    def run(self):
        self.compute_changes()

    def print(self):
        self.write_summary_rst()
        self.write_summary_js()


def aggregate(source_path, summary_path):
    logger.info("Computing aggregated CHANGES and summarizing in reStructuredText format")
    ca = ChangesAggregator(source_path, summary_path)
    logger.info(f"Input path: {ca.project_path}")
    logger.info(f"Output path: {ca.summary_path}")
    ca.run()
    logger.info("Ready: Found %s changes in %s projects" % (len(ca.changes), len(ca.projects)))
    ca.print()



import typing as t
from pathlib import Path

import click
from click_aliases import ClickAliasedGroup

from rapporto.option import format_option
from rapporto.source.github.actions import GitHubActionsReport
from rapporto.source.github.activity import GitHubActivityReport
from rapporto.source.github.attention import GitHubAttentionReport
from rapporto.source.github.backup import GitHubBackup
from rapporto.source.github.model import GitHubInquiry, GitHubMultiRepositoryInquiry, GitHubOptions
from rapporto.util import to_mrkdwn

organization_option = click.option("--organization", "--org", type=str, required=False)
author_option = click.option("--author", type=str, required=False)
when_option = click.option("--when", type=str, required=False, help="Time interval")
repository_option = click.option(
    "--repository",
    type=str,
    required=True,
    help="GitHub repository, single or path to file",
)


@click.group(cls=ClickAliasedGroup)
@click.pass_context
def cli(ctx: click.Context):
    """
    Harvest information from GitHub.
    """
    pass


@cli.command(aliases=["ppp"])
@organization_option
@author_option
@when_option
@format_option
def activity(
    organization: t.Optional[str] = None,
    author: t.Optional[str] = None,
    when: t.Optional[str] = None,
    format_: t.Optional[str] = None,
):
    """
    Activities of individual authors.
    """
    inquiry = GitHubInquiry(organization=organization, author=author, updated=when)
    report = GitHubActivityReport(inquiry=inquiry)
    print_output(report, format_)


def print_output(report, format_):
    if format_ == "markdown":
        print(report.markdown)
    elif format_ == "mrkdwn":
        print(to_mrkdwn(report.markdown))
    else:
        raise NotImplementedError(f"Unknown output format: {format_}")


@cli.command(aliases=["ci"])
@repository_option
@when_option
@format_option
def actions(
    repository: t.Union[str, Path],
    when: t.Optional[str] = None,
    format_: t.Optional[str] = None,
):
    """
    CI/GHA failures.
    """
    options = GitHubOptions().add_repos(repository)
    inquiry = GitHubMultiRepositoryInquiry(repositories=options.repositories, created=when)
    report = GitHubActionsReport(inquiry=inquiry)
    print_output(report, format_)


@cli.command(aliases=["att"])
@organization_option
@when_option
@format_option
def attention(
    organization: t.Optional[str] = None,
    when: t.Optional[str] = None,
    format_: t.Optional[str] = None,
):
    """
    Important items that deserve attention.
    """
    inquiry = GitHubInquiry(organization=organization, updated=when)
    report = GitHubAttentionReport(inquiry=inquiry)
    print_output(report, format_)


@cli.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    )
)
@click.pass_context
def backup(ctx: click.Context):
    """
    Backup GitHub project.
    """
    backup = GitHubBackup()
    try:
        backup.run(ctx.args)
    except Exception as ex:
        click.echo(f"ERROR: {ex}", err=True)
        raise SystemExit(2) from ex



import click

from rapporto.animate.git import render


@click.group()
@click.pass_context
def cli(
    ctx: click.Context,
):
    """
    Animate time series information from different sources.
    """
    pass


@cli.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
@click.pass_context
def git(ctx: click.Context, args):
    """
    Render video of VCS repository commit history.
    """
    render(list(args))



from rapporto.util import sanitize_title, week_to_day_range


def test_sanitize_title():
    assert sanitize_title("[Foo] Hotzenplotz") == "⎡Foo⎦ Hotzenplotz"


def test_week_to_day_range():
    assert len(week_to_day_range("2025W09", skip_the_future=False)) == 7



import datetime as dt
import io
import typing as t

import attr
import yaml
from attrs import define

from rapporto.source.github.actions import GitHubActionsReport
from rapporto.source.github.attention import GitHubAttentionReport
from rapporto.source.github.model import GitHubInquiry, GitHubMultiRepositoryInquiry, GitHubOptions
from rapporto.util import week_to_day_range


@define
class DailyItem:
    """
    Represent a single item of daily recurring report information.
    """

    type: str
    day: str
    markdown: str


@define
class WeeklyItem:
    """
    Represent a multiple weekly items of daily recurring report information.
    """

    type: str
    week: str
    items: t.List[DailyItem] = attr.field(factory=list)


@define
class ReportOptions:
    """
    Composite report options.
    """

    output_format: str = "markdown"


@define
class ReportBase:
    """
    Common base class for reports.
    """

    def to_dict(self) -> t.Dict[str, t.Any]:
        return attr.asdict(self)

    def data_items(self) -> t.List[t.Union[DailyItem, DailyItem]]:
        return self.to_dict()["data"]

    def to_yaml(self) -> str:
        return yaml.dump(self.to_dict(), sort_keys=False)

    def to_json(self) -> str:
        raise NotImplementedError("Needs special support by `cattrs`, see Zyp/Tikray")
        # return json.dumps(yaml.load(self.to_yaml(), yaml.UnsafeLoader), sort_keys=False, indent=2)
        # return json.dumps(munchify(self.to_dict()), sort_keys=False, indent=2)

    @property
    def markdown(self):
        return self.to_markdown()

    def to_markdown(self) -> str:
        buffer = io.StringIO()
        for item in self.data_items():
            buffer.write(item.markdown)
            buffer.write("\n\n")
        return buffer.getvalue()

    def render(self, format_: str):
        if format_ in ["markdown", "md"]:
            return self.markdown
        elif format_ == "yaml":
            return self.to_yaml()
        elif format_ == "json":
            return self.to_json()
        else:
            raise NotImplementedError(f"Unknown format: {format_}")


@define
class DailyReport(ReportBase):
    """
    Produce a daily recurring report.
    """

    day: str
    github_options: GitHubOptions
    report_options: ReportOptions
    items: t.List[DailyItem] = attr.field(factory=list)

    def __attrs_post_init__(self):
        """
        Use current datetime by default.
        """
        if self.day is None:
            self.day = dt.datetime.now().strftime("%Y-%m-%d")

    def process(self):
        """
        Generate set of reports across different domains or topics.
        """
        self.github_actions()
        self.github_attention()

    def github_actions(self):
        """
        CI workflow run failures on GitHub.
        """
        # TODO: Use `TimeIntervalParser`.
        created = f"{self.day}..{self.day}"
        inquiry = GitHubMultiRepositoryInquiry(
            repositories=self.github_options.repositories, created=created
        )
        report = GitHubActionsReport(inquiry=inquiry)
        self.items.append(DailyItem(type="github-actions", day=self.day, markdown=report.markdown))

    def github_attention(self):
        """
        Items on GitHub that deserve your attention.
        """
        # TODO: Use `TimeIntervalParser`.
        updated = f"{self.day}..{self.day}"
        inquiry = GitHubInquiry(organization=self.github_options.organization, updated=updated)
        report = GitHubAttentionReport(inquiry=inquiry)
        self.items.append(
            DailyItem(type="github-attention", day=self.day, markdown=report.markdown)
        )

    def to_dict(self):
        return {
            "meta": {"day": self.day},
            "data": self.items,
        }


@define
class WeeklyReport(ReportBase):
    """
    Produce a weekly recurring report.
    """

    week: str
    github_options: GitHubOptions
    report_options: ReportOptions
    dailies: t.List[DailyReport] = attr.field(factory=list)

    SKIP_THE_FUTURE = True

    def __attrs_post_init__(self):
        """
        Use current calendar week by default.
        """
        if self.week is None:
            self.week = dt.datetime.now().strftime("%YW%V")

    @property
    def days(self) -> t.List[str]:
        """
        Enumerate days of designated week.

        TODO: Refactor to Aika.
        """
        return week_to_day_range(self.week, skip_the_future=self.SKIP_THE_FUTURE)

    def process(self):
        """
        Create all daily reports.
        """
        for day in self.days:
            report = DailyReport(
                day=day, github_options=self.github_options, report_options=self.report_options
            )
            report.process()
            self.dailies.append(report)

    def to_dict(self):
        return {
            "meta": {"week": self.week},
            "data": self.dailies,
        }



import logging

import click

from rapporto.option import (
    format_option,
    github_organization_option,
    github_repository_option,
)
from rapporto.report.model import DailyReport, ReportOptions, WeeklyReport
from rapporto.source.github.model import GitHubOptions

logger = logging.getLogger(__name__)


@click.group()
@github_organization_option
@github_repository_option
@format_option
@click.pass_context
def cli(ctx: click.Context, github_organization: str, github_repository: str, format_: str):
    """
    Generate reports.
    """
    ctx.meta["github_options"] = GitHubOptions(organization=github_organization).add_repos(
        github_repository
    )
    ctx.meta["report_options"] = ReportOptions(output_format=format_)


@cli.command()
@click.option("--day", type=str, required=False, help="Day in ISO format, e.g. 2025-03-03")
@click.pass_context
def daily(ctx: click.Context, day: str):
    """
    Daily report.
    """
    report = DailyReport(
        day=day,
        github_options=ctx.meta["github_options"],
        report_options=ctx.meta["report_options"],
    )
    report.process()
    print(report.render(report.report_options.output_format))


@cli.command()
@click.option("--week", type=str, required=False, help="Calendar week in ISO format, e.g. 2025W03")
@click.pass_context
def weekly(ctx: click.Context, week: str):
    """
    Weekly report.
    """
    report = WeeklyReport(
        week=week,
        github_options=ctx.meta["github_options"],
        report_options=ctx.meta["report_options"],
    )
    report.process()
    print(report.render(report.report_options.output_format))



import logging
import os
import urllib.parse

import requests_cache

logger = logging.getLogger(__name__)


def repository_name(url: str, with_org: bool = False) -> str:
    """
    Compute GitHub repository name from full URL.

    https://github.com/tech-writing/rapporto
    https://github.com/tech-writing/rapporto/issues/19
    https://api.github.com/repos/tech-writing/rapporto
    """
    if "api.github.com" in url:
        url = url.replace("/repos", "")
    uri = urllib.parse.urlparse(url)
    parts = uri.path.split("/")
    if with_org:
        return parts[1] + "/" + parts[2]
    else:
        return parts[2]


class GitHubHttpClient:
    session = requests_cache.CachedSession(backend="sqlite", expire_after=3600)
    if "GH_TOKEN" in os.environ:
        session.headers.update({"Authorization": f"Bearer {os.getenv('GH_TOKEN')}"})
    else:
        logger.warning("GH_TOKEN not defined. This will exhaust the rate limit quickly.")



import logging

import click

from pueblo_goof.slack.model import slack_api_token_option
from rapporto.source.slack.core import SlackThreadExporter

logger = logging.getLogger(__name__)


@click.group()
@slack_api_token_option
@click.pass_context
def cli(ctx: click.Context, slack_token: str):
    """
    Harvest information from Slack.
    """
    if not slack_token:
        raise click.UsageError(
            "Missing option '--slack-token' or environment variable 'SLACK_TOKEN'."
        )
    ctx.meta.update({"slack_token": slack_token})


@cli.command()
@click.argument("url", type=str, required=True)
@click.pass_context
def export(ctx: click.Context, url: str):
    """
    Export a Slack thread into Markdown format.
    """
    exporter = SlackThreadExporter(ctx.meta["slack_token"])
    exporter.export_thread(url)



import dataclasses
import datetime as dt
import logging
import typing as t
import urllib.parse
from abc import abstractmethod
from collections import OrderedDict
from enum import Enum
from pathlib import Path

import attr
import dateparser
from aika import TimeInterval, TimeIntervalParser
from attrs import define

logger = logging.getLogger(__name__)


@define
class GitHubOptions:
    """
    Options for GitHub.
    """

    organization: t.Optional[str] = None
    repositories: t.List[str] = attr.field(factory=list)

    def add_repos(self, repos: t.Union[str, Path]) -> "GitHubOptions":
        if repos is None:
            return self
        path = Path(repos)
        if path.exists():
            self.add_repositories_file(path)
        elif isinstance(repos, str):
            self.add_repository(repos)
        else:
            raise TypeError("repos must be a str or Path")
        return self

    def add_repository(self, name: str):
        self.repositories.append(name)

    def add_repositories_file(self, path: Path):
        self.repositories += path.read_text().splitlines()


@dataclasses.dataclass
class GitHubInquiry:
    organization: t.Optional[str] = None
    updated: t.Optional[str] = None
    author: t.Optional[str] = None


@dataclasses.dataclass
class GitHubMultiRepositoryInquiry:
    repositories: t.List[str]
    created: t.Optional[str] = None


class QType(Enum):
    API = 1
    HTML = 2


class QKind(Enum):
    ISSUE = 1
    PULLREQUEST = 2


class GitHubQueryBuilder:
    template_api = (
        "https://api.github.com/search/issues?q={query}&per_page=100&sort=created&order=asc"
    )
    template_html = "https://github.com/search?q={query}&per_page=100&s=created&o=asc"

    def __init__(self, inquiry: GitHubInquiry):
        self.inquiry = inquiry
        self.type: t.Optional[QType] = None
        self.kind: t.Optional[QKind] = None
        self.constraints: t.List[str] = []
        self.query()

    def add(self, field: str, value: t.Optional[str] = None, is_list: bool = False):
        if value is not None:
            if is_list:
                for item in value.split(","):
                    self.add(field, item.strip())
                return
            self.constraints.append(f"{field}:{value}")

    @property
    def expression(self):
        return " ".join(self.constraints)

    @abstractmethod
    def query(self):
        raise NotImplementedError("Needs to be implemented")

    @property
    def timeinterval(self) -> TimeInterval:
        return timeinterval(self.inquiry.updated)

    @property
    def timerange(self) -> str:
        timerange_user = self.inquiry.updated
        timerange_effective = self.timeinterval.githubformat()
        logger.info(f'Using timerange: user="{timerange_user}" effective="{timerange_effective}"')
        return timerange_effective

    @property
    def query_issues(self):
        return f"{self.expression} is:issue"

    @property
    def query_pulls(self):
        return f"{self.expression} is:pr"

    def issue(self):
        self.kind = QKind.ISSUE
        return self

    def pr(self):
        self.kind = QKind.PULLREQUEST
        return self

    def api(self):
        self.type = QType.API
        return self

    def html(self):
        self.type = QType.HTML
        return self

    def url(self):
        if self.type == QType.API:
            template = self.template_api
        elif self.type == QType.HTML:
            template = self.template_html
        else:
            raise NotImplementedError("Unknown type: Only API and HTML are supported")
        if self.kind == QKind.ISSUE:
            query = self.query_issues
        elif self.kind == QKind.PULLREQUEST:
            query = self.query_pulls
        else:
            raise NotImplementedError("Unknown kind: Only ISSUE and PULLREQUEST are supported")
        return template.format(query=urllib.parse.quote(query))


@define
class GitHubSearch:
    session: t.Any
    query_builder: GitHubQueryBuilder
    issues_api: str
    pulls_api: str
    issues_html: str
    pulls_html: str

    @classmethod
    def with_query_builder(cls, session, query_builder: GitHubQueryBuilder):
        return cls(
            session=session,
            query_builder=query_builder,
            issues_api=query_builder.issue().api().url(),
            pulls_api=query_builder.pr().api().url(),
            issues_html=query_builder.issue().html().url(),
            pulls_html=query_builder.pr().html().url(),
        )

    def issues_and_prs(self):
        items = []

        response = self.session.get(self.issues_api)
        response.raise_for_status()
        items += response.json()["items"]

        response = self.session.get(self.pulls_api)
        response.raise_for_status()
        items += response.json()["items"]

        return items


@dataclasses.dataclass()
class MarkdownContent:
    labels: t.OrderedDict[str, str] = dataclasses.field(default_factory=OrderedDict)
    content: t.Dict[str, t.List[str]] = dataclasses.field(default_factory=dict)

    def add(self, section, content):
        self.content.setdefault(section, [])
        self.content[section].append(content)

    def render_section(self, section) -> t.Optional[str]:
        if section not in self.content:
            return None
        label = self.labels.get(section, section)
        body = "\n".join(self.content[section])
        return f"\n## {label}\n{body}"

    def render(self):
        sections = []
        for section in self.labels.keys():
            if markdown := self.render_section(section):
                sections.append(markdown)
        return "\n".join(sections)


def timeinterval(when: t.Optional[str] = None) -> TimeInterval:
    if when is None:
        return TimeInterval(dt.datetime.today())

    # TODO: Add edge case to Aika. -- https://github.com/panodata/aika/issues/112
    if when and ".." in when:
        parts = when.split("..")
        return TimeInterval(start=dateparser.parse(parts[0]), end=dateparser.parse(parts[1]))
    # TODO: Let optionally Aika NOT set the `end` part of the interval to NOW.
    tr = TimeIntervalParser()
    return tr.parse(when)



"""
Render video of VCS repository commit history using Gource, with audio.

Prerequisites::

    apt-get install --yes ffmpeg gource mp3wrap
    brew install ffmpeg gource mp3wrap

Synopsis::

    rapporto animate git \
        --name acme \
        --path ~/dev/sandbox/acme \
        --audio '/home/foobar/music/Beastie boys/Suco De Tangerina.mp3'

References:
- https://gource.io/
- https://github.com/acaudwell/Gource
- zt.manticore.ext.gource
"""

# Relax linter.
# `subprocess` call: check for execution of untrusted input
# Starting a process with a shell, possible injection detected
# ruff: noqa: S603,S605
import math
import os
import re
import shlex
import shutil
import subprocess
import sys
import time


class GourceRenderer:
    """Renders project history using 'gource'."""

    def __init__(
        self,
        source_path,
        target_path,
        overwrite=False,
        audio=None,
        start_date=None,
        stop_date=None,
        time_lapse=False,
    ):
        self.project_path = source_path
        self.output_path = target_path

        self.start_date = start_date
        self.stop_date = stop_date

        # configuration data
        self.gource_cmd_tpl = """
            gource \\
                --title "%(title)s" --key \\
                --viewport 1280x720 \\
                --multi-sampling \\
                --hide bloom  \\
                --output-ppm-stream - \\
%(date_options)s  \\
%(speed_options)s  \\
                --stop-at-end \\
                --max-user-speed 250 \\
                --user-scale 1.5 \\
                --user-font-size 18 \\
                %(path)s \\
        """

        # --disable-auto-rotate \\

        # options
        self.overwrite = overwrite
        self.audio = audio
        self.time_lapse = time_lapse

    def get_gource_command(self, path: str, title: str):
        date_options = ""
        if self.start_date is not None:
            date_options += f"                --start-date '{self.start_date}' \\\n"
        if self.stop_date is not None:
            date_options += f"                --stop-date '{self.stop_date}' \\\n"
        else:
            date_options += "                --stop-at-end \\\n"
        # --stop-at-time 1

        speed_options = "                --seconds-per-day 5 --time-scale 1.5 \\\n"
        if self.time_lapse:
            speed_options = "                --seconds-per-day 2.5 --time-scale 2 \\\n"
        speed_options += "                --file-idle-time 20 --max-file-lag 2.5 \\\n"

        gource_options = {
            "path": path,
            "title": title,
            "date_options": date_options,
            "speed_options": speed_options,
        }
        command = self.gource_cmd_tpl % gource_options
        return command

    def choose_background_song(self):
        # TODO: enhance song picker (e.g. random or mapped selection from a directory)
        if self.audio and os.path.isfile(self.audio):
            return self.audio
        elif os.path.isfile(os.environ.get("GOURCE_AUDIO", "")):
            return os.environ.get("GOURCE_AUDIO")
        return None

    def process_project(self, path: str, name: str):
        print("=" * 42)
        print("Processing project '%s'" % name)
        print("=" * 42)

        video_file = self.create_video(path, self.output_path, name)
        if not video_file:
            print("ERROR: video could not be created")
            return False

        mi = MediaInfo(video_file)
        if "video" in mi.get_streams():
            print("INFO: Video:    ", video_file)
            print("INFO: Duration: ", mi.duration)

        else:
            print("ERROR: video '%s' could not be recorded" % video_file)
            os.unlink(video_file)
            video_file = None
            return False

        return True

    def create_video(self, project_path, video_path, video_filename):
        vr = VideoRecorder(video_path, video_filename)
        if vr.exists() and not self.overwrite:
            print("INFO: Video exists and --overwrite is not given, will skip further processing.")
            return vr.get_video_file()

        print("-" * 42)
        print("Creating video '%s'" % vr.get_video_file())
        cmd = (
            self.get_gource_command(path=project_path, title=video_filename)
            + " | \\"
            + vr.get_command()
        )
        print("command:", cmd)
        print("-" * 42)

        if run_command(cmd):
            video_file = vr.get_video_file()

            audio_file = self.choose_background_song()
            if audio_file:
                mixer = VideoAudioMixer(video_file, audio_file)
                mixer.extend_audio()
                mixer.run()

            return video_file
        return None


class MediaInfo:
    """
    Duration: 00:00:01.73, start: 0.000000, bitrate: 662 kb/s
      Stream #0.0(und): Video: h264, yuv420p, 1024x768, 651 kb/s, 60 fps, 59.94 tbr, 60 tbn, 120 tbc
    """

    def __init__(self, mediafile):
        self.mediafile = mediafile
        self.raw = ""
        self.duration = ""
        self.streams = []
        self.read_info()
        self.parse_info()

    def read_info(self):
        cmd = "ffprobe -i '%s'" % self.mediafile
        print("MediaInfo ffmpeg command:", cmd)
        output = subprocess.check_output(shlex.split(cmd), stderr=subprocess.STDOUT).decode("utf-8")
        # print("MediaInfo output:", output)
        self.raw = output

    def parse_info(self):
        r_duration = re.compile(".*Duration: ([.:0-9]+)")
        r_stream = re.compile(".*Stream .*: (Video|Audio)")
        for line in self.raw.split("\n"):
            m = r_duration.match(line)
            if m:
                self.duration = m.group(1)
            m = r_stream.match(line)
            if m:
                self.streams.append(m.group(1).lower())

    def get_duration(self):
        return self.duration

    def get_streams(self):
        return self.streams


class VideoRecorder:
    def __init__(self, video_path, video_name):
        self.video_path = video_path
        self.video_name = video_name
        self.video_file = self.get_video_file()

    def get_command(self):
        arguments = self.__dict__.copy()

        # Some remarks about "ffmpeg" options:
        #   - The encoder 'aac' is experimental but experimental codecs are not enabled,
        #     add '-strict -2' if you want to use it.
        command = (
            f"""
            ffmpeg -y \\
                -r 60 \\
                -vcodec ppm -f image2pipe -i - \\
                -vcodec libx264 -pix_fmt yuv420p -preset medium -threads 0 -strict -2 \\
                "{self.video_file}" \\
            """
            % arguments
        )
        return command

    def get_video_file(self):
        return os.path.join(self.video_path, self.video_name + self.get_extension())

    def get_extension(self):
        return ".mp4"

    def exists(self):
        return os.path.exists(self.get_video_file())


class VideoAudioMixer:
    def __init__(self, video_file, audio_file):
        self.video_file = video_file
        self.audio_file = audio_file
        self.duration = None

    def run(self):
        extension = os.path.splitext(self.video_file)[1]
        tmpfile = self.video_file + ".tmp" + extension
        cmd = "ffmpeg -y -i '%s' -i '%s' -vcodec copy -acodec copy -async 1 -shortest '%s'" % (
            self.video_file,
            self.audio_file,
            tmpfile,
        )
        print("VideoAudioMixer ffmpeg command:", cmd)
        output = subprocess.check_output(shlex.split(cmd), stderr=subprocess.STDOUT)
        print("VideoAudioMixer ffmpeg output:", output)
        shutil.move(tmpfile, self.video_file)

    def extend_audio(self):
        video_info = MediaInfo(mediafile=self.video_file)
        audio_info = MediaInfo(mediafile=self.audio_file)
        loops = self.get_audio_loops(video_info.duration, audio_info.duration)
        print(f"Repeating audio {loops} times to match length of video")
        self.audio_file = self.loop_audio(times=loops)

    def loop_audio(self, times=2):
        # TODO: remove looped audio file after usage

        name = os.path.basename(self.audio_file)
        audio_files = ('"%s" ' % self.audio_file) * times
        mp3wrap_file = "/tmp/tmp_%s" % name

        if not times or times <= 1:
            shutil.copy(self.audio_file, mp3wrap_file)
            return mp3wrap_file

        # WTF?
        mp3wrap_file_real = mp3wrap_file.replace(".mp3", "_MP3WRAP.mp3")
        if os.path.exists(mp3wrap_file_real):
            os.unlink(mp3wrap_file_real)

        mp3wrap_command = 'mp3wrap "%s" %s' % (mp3wrap_file, audio_files)
        if run_command(mp3wrap_command):
            return mp3wrap_file_real
        return None

    @classmethod
    def get_audio_loops(cls, video_duration, audio_duration):
        """
        d1 = "00:03:16.18"
        d2 = "00:00:06.43"
        >>> get_audio_loops(d1, d2)
        33
        """
        factor = cls.duration_to_seconds(video_duration) / cls.duration_to_seconds(audio_duration)
        loops = math.ceil(factor)
        return loops

    @classmethod
    def duration_to_seconds(cls, duration):
        time_struct = time.strptime(duration, "%H:%M:%S.%f")
        seconds = time_struct.tm_hour * 3600 + time_struct.tm_min * 60 + time_struct.tm_sec
        return seconds


def run_command(command):
    returncode = os.system(command)
    if returncode == 0:
        return True
    else:
        print("ERROR while executing command '%s'" % command)
        return False


def render(args):
    from optparse import OptionParser

    parser = OptionParser()
    parser.add_option("-p", "--path", dest="path", help="path to vcs repository")
    parser.add_option("-o", "--outdir", dest="outdir", help="path to output directory")
    parser.add_option(
        "-n",
        "--name",
        dest="name",
        help="project name (output video basename w/o extension) [optional]",
    )
    parser.add_option("-a", "--audio", dest="audio", type="str", help="path to audio file")
    parser.add_option(
        "-O",
        "--overwrite",
        dest="overwrite",
        action="store_true",
        help="whether to overwrite video files",
    )
    parser.add_option("--start-date", dest="start_date", help="Start at a date and optional time")
    parser.add_option("--stop-date", dest="stop_date", help="Stop at a date and optional time")
    parser.add_option(
        "-t", "--time-lapse", dest="time_lapse", action="store_true", help="run in time-lapse mode"
    )
    (options, args) = parser.parse_args(args=args)

    # Sanity checks.
    if not options.path:
        print("ERROR: Option '--path' is mandatory!")
        sys.exit(1)

    options.path = os.path.abspath(options.path)
    if not os.path.isdir(options.path):
        print("ERROR: Directory '%s' does not exist" % options.path)
        sys.exit(1)

    # Defaults.
    if not options.outdir:
        options.outdir = os.path.abspath(os.path.curdir)

    if not options.name:
        options.name = os.path.basename(options.path)

    print(
        "Rendering project history of single project '%s <%s>' using 'gource'"
        % (options.name, options.path)
    )
    source_path = options.path
    target_path = options.outdir
    gr = GourceRenderer(
        source_path,
        target_path,
        overwrite=options.overwrite,
        audio=options.audio,
        start_date=options.start_date,
        stop_date=options.stop_date,
        time_lapse=options.time_lapse,
    )
    gr.process_project(path=options.path, name=options.name)



import typing as t
from collections import OrderedDict
from operator import attrgetter

from munch import Munch, munchify
from tqdm import tqdm

from rapporto.source.github.model import (
    GitHubInquiry,
    GitHubQueryBuilder,
    GitHubSearch,
    MarkdownContent,
)
from rapporto.source.github.util import GitHubHttpClient, repository_name
from rapporto.util import goosefeet, sanitize_title


class GitHubAttentionQueryBuilder(GitHubQueryBuilder):
    """
    Find all open issues and pull requests with labels "bug" or "important".
    """

    labels: t.ClassVar[t.List[str]] = [
        "bug",  # GitHub standard.
        "important",
        "incident",
        "stale",
        "type-bug",  # CPython
        "type-crash",  # CPython
        "type: bug",
        "type: incident",
    ]

    def query(self):
        self.add("org", self.inquiry.organization)
        self.add("updated", self.timeinterval.githubformat())
        self.add("label", ",".join(map(goosefeet, self.labels)))
        # self.add("state", "open")


class GitHubAttentionReport:
    """
    Report about important items that deserve your attention, bugs first.

    Find all issues and pull requests with labels "bug" or "important".
    """

    label_section_map: t.ClassVar[t.OrderedDict[str, str]] = OrderedDict(
        bug="Bugs",
        incident="Incidents",
        important="Important",
        stale="Stale",
        others="Others",
    )

    label_aliases: t.ClassVar[t.Dict[str, t.List[str]]] = {
        "bug": ["type-bug", "type-crash", "type: bug"],
        "incident": ["type: incident"],
    }

    def __init__(self, inquiry: GitHubInquiry):
        self.inquiry = inquiry
        self.session = GitHubHttpClient.session
        self.search = GitHubSearch.with_query_builder(
            self.session, GitHubAttentionQueryBuilder(inquiry=inquiry)
        )

    @property
    def items(self):
        """
        Return GitHub issues and PRs in scope of search constraints.
        """
        items = self.search.issues_and_prs()
        return sorted(munchify(items), key=attrgetter("created_at"))

    def has_relevant_label(self, item) -> t.Optional[Munch]:
        """
        Whether the given item includes a relevant label.
        """
        for label in item.labels:
            label_name = label.name.lower()
            for label_key in self.label_section_map.keys():
                label_key = label_key.lower()
                if label_name == label_key or label_name in self.label_aliases.get(label_key, []):
                    # It's easier for the downstream renderer when using canonical category labels.
                    label.category = label_key
                    return label
        return None

    @property
    def markdown(self):
        """
        Render report in Markdown format.
        """
        mdc = MarkdownContent(labels=self.label_section_map)
        seen = {}
        for item in tqdm(self.items, leave=False):
            is_closed = item.state == "closed"
            title = sanitize_title(
                f"{repository_name(item.repository_url, with_org=True)}: {item.title}"
            )
            link = f"[{title}]({item.html_url})"
            if is_closed:
                line = f"- ~{link}~"
            else:
                line = f"- {link}"
            # line = f"- {link} {', '.join(labels)}"
            if label := self.has_relevant_label(item):
                if item.html_url in seen:
                    continue
                seen[item.html_url] = True
                mdc.add(label.category, line)
            else:
                mdc.add("others", line)

        body = mdc.render()
        if not body:
            easy_day = "[easy day](https://www.youtube.com/watch?v=neg7yhoZgdw)"
            body = (
                f"_Apparently, nothing significant happened yet. "
                f"Please also have a look at previous editions. "
                f"Otherwise, have an {easy_day}._"
            )

        link_issues = f"[Issues]({self.search.issues_html})"
        link_pulls = f"[Pull requests]({self.search.pulls_html})"

        return f"""
# Attention report {self.inquiry.updated or ""}

A report about important items that deserve your attention, bugs first.
- Time range: {self.search.query_builder.timeinterval.githubformat() or "n/a"}
- Details: {link_issues}, {link_pulls}

{body}
        """.strip()



import logging

import click

from pueblo_goof.slack.conversation import SlackConversation
from pueblo_goof.slack.model import SlackOptions, slack_api_token_option, slack_channel_option
from pueblo_goof.util import Zapper
from rapporto.notify.slack import SlackWeekly
from rapporto.option import github_organization_option, github_repository_option
from rapporto.report.model import ReportOptions
from rapporto.source.github.model import GitHubOptions

logger = logging.getLogger(__name__)


@click.group()
@github_organization_option
@github_repository_option
@slack_api_token_option
@slack_channel_option
@click.pass_context
def cli(
    ctx: click.Context,
    github_organization: str,
    github_repository: str,
    slack_token: str,
    slack_channel: str,
):
    """
    Notify humans and machines.
    """
    if not slack_token:
        raise click.UsageError(
            "Missing option '--slack-token' or environment variable 'SLACK_TOKEN'."
        )
    ctx.meta["github_options"] = GitHubOptions(organization=github_organization).add_repos(
        github_repository
    )
    ctx.meta["slack_options"] = SlackOptions(token=slack_token, channel=slack_channel)
    ctx.meta["report_options"] = ReportOptions(output_format="markdown")


@cli.command()
@click.option("--week", type=str, required=False, help="Calendar week")
@click.option("--zap", type=str, required=False, help="Zap message again")
@click.pass_context
def weekly(ctx: click.Context, week: str, zap: str):
    """
    Weekly report converged into Slack thread.
    """

    github_options: GitHubOptions = ctx.meta["github_options"]
    report_options: ReportOptions = ctx.meta["report_options"]
    slack_options: SlackOptions = ctx.meta["slack_options"]

    conversation = SlackConversation(options=slack_options)
    zapper = Zapper(when=zap, action=conversation.delete)

    section = SlackWeekly(
        week=week,
        github_options=github_options,
        report_options=report_options,
        conversation=conversation,
    )
    section.refresh()

    try:
        zapper.process()
    except Exception as e:
        raise click.ClickException(f"The 'zap' operation failed: {e}") from e



import click
from click_aliases import ClickAliasedGroup

from rapporto.animate.cli import cli as animate_cli
from rapporto.notify.cli import cli as notify_cli
from rapporto.report.cli import cli as report_cli
from rapporto.source.changes.cli import cli as changes_cli
from rapporto.source.github.cli import cli as github_cli
from rapporto.source.opsgenie.cli import cli as opsgenie_cli
from rapporto.source.slack.cli import cli as slack_cli
from rapporto.util import setup_logging


@click.group(cls=ClickAliasedGroup)
@click.option("--verbose", is_flag=True, required=False, help="Turn on logging")
@click.option("--debug", is_flag=True, required=False, help="Turn on logging with debug level")
@click.version_option()
@click.pass_context
def cli(ctx: click.Context, verbose: bool, debug: bool):
    setup_logging(verbose=verbose)


cli.add_command(github_cli, "github", aliases=["gh"])
cli.add_command(opsgenie_cli, "opsgenie")
cli.add_command(slack_cli, "slack")
cli.add_command(notify_cli, "notify")
cli.add_command(report_cli, "report")
cli.add_command(animate_cli, "animate")
cli.add_command(changes_cli, "changes")
